@@ -261,11 +261,11 @@ var envelopeTests = []struct {
 			Date:      t,
 			Subject:   "Hello World!",
 			From:      []*Address{addrTests[0].addr},
-			Sender:    []*Address{},
-			ReplyTo:   []*Address{},
-			To:        []*Address{},
-			Cc:        []*Address{},
-			Bcc:       []*Address{},
+			Sender:    nil,
+			ReplyTo:   nil,
+			To:        nil,
+			Cc:        nil,
+			Bcc:       nil,
 			InReplyTo: "42@example.org",
 			MessageId: "43@example.org",
 		},
@@ -273,11 +273,11 @@ var envelopeTests = []struct {
 			"Tue, 10 Nov 2009 23:00:00 -0600",
 			"Hello World!",
 			[]interface{}{addrTests[0].fields},
-			[]interface{}{},
-			[]interface{}{},
-			[]interface{}{},
-			[]interface{}{},
-			[]interface{}{},
+			nil,
+			nil,
+			nil,
+			nil,
+			nil,
 			"42@example.org",
 			"43@example.org",
 		},
@@ -372,6 +372,24 @@ func TestAddress_Format(t *testing.T) {
 	}
 }
 
+func TestEmptyAddress(t *testing.T) {
+	fields := []interface{}{nil, nil, nil, nil}
+	addr := &Address{}
+	err := addr.Parse(fields)
+	if err == nil {
+		t.Error("A nil address did not return an error")
+	}
+}
+
+func TestEmptyGroupAddress(t *testing.T) {
+	fields := []interface{}{nil, nil, "undisclosed-recipients", nil}
+	addr := &Address{}
+	err := addr.Parse(fields)
+	if err == nil {
+		t.Error("An empty group did not return an error when parsed as address")
+	}
+}
+
 func TestAddressList(t *testing.T) {
 	fields := make([]interface{}, len(addrTests))
 	addrs := make([]*Address, len(addrTests))
@@ -391,6 +409,16 @@ func TestAddressList(t *testing.T) {
 	}
 }
 
+func TestEmptyAddressList(t *testing.T) {
+	addrs := make([]*Address, 0)
+
+	gotFields := FormatAddressList(addrs)
+	if !reflect.DeepEqual(gotFields, nil) {
+		t.Error("Invalid address list fields: got", gotFields, "but expected nil")
+	}
+}
+
+
 var paramsListTest = []struct {
 	fields []interface{}
 	params map[string]string
@@ -479,12 +507,12 @@ var bodyStructureTests = []struct {
 			Size:        42,
 			Lines:       67,
 			Envelope: &Envelope{
-				From:    []*Address{},
-				Sender:  []*Address{},
-				ReplyTo: []*Address{},
-				To:      []*Address{},
-				Cc:      []*Address{},
-				Bcc:     []*Address{},
+				From:    nil,
+				Sender:  nil,
+				ReplyTo: nil,
+				To:      nil,
+				Cc:      nil,
+				Bcc:     nil,
 			},
 			BodyStructure: &BodyStructure{
 				Params: map[string]string{},
@@ -584,6 +612,36 @@ func TestBodyStructure_Parse(t *testing.T) {
 	}
 }
 
+func TestBodyStructure_Parse_uppercase(t *testing.T) {
+	fields := []interface{}{
+		"APPLICATION", "PDF", []interface{}{"NAME", "Document.pdf"}, nil, nil,
+		"BASE64", RawString("4242"), nil,
+		[]interface{}{"ATTACHMENT", []interface{}{"FILENAME", "Document.pdf"}},
+		nil, nil,
+	}
+
+	expected := &BodyStructure{
+		MIMEType:          "application",
+		MIMESubType:       "pdf",
+		Params:            map[string]string{"name": "Document.pdf"},
+		Encoding:          "base64",
+		Size:              4242,
+		Extended:          true,
+		MD5:               "",
+		Disposition:       "attachment",
+		DispositionParams: map[string]string{"filename": "Document.pdf"},
+		Language:          nil,
+		Location:          []string{},
+	}
+
+	bs := &BodyStructure{}
+	if err := bs.Parse(fields); err != nil {
+		t.Errorf("Cannot parse: %v", err)
+	} else if !reflect.DeepEqual(bs, expected) {
+		t.Errorf("Invalid body structure: got \n%+v\n but expected \n%+v", bs, expected)
+	}
+}
+
 func TestBodyStructure_Format(t *testing.T) {
 	for i, test := range bodyStructureTests {
 		fields := test.bodyStructure.Format()
